#include <fstream>
#include <sstream>
#include "Rule.h"

/************************************************
  version
************************************************/
void version()
{
	cout << "----------------------------------------\n"
		    " reverser v0.2\n"
			" 2010/03/02 - 2010/03/02\n"
			" (c) 2008-2010 Yang Liu\n"
			"----------------------------------------\n";
}

/************************************************
  dump 
************************************************/
void dump(const char* fileName,
		  const map<Rule, float>& bpMap)
{
	ofstream out(fileName);

	map<Rule, float>::const_iterator iter;

	for (iter = bpMap.begin(); iter != bpMap.end(); iter++)
	{
		out << iter->first
			<< "||| "
			<< iter->second
			<< endl;
	}
}

/************************************************
  main function
************************************************/
int main(int argc, char** argv)
{
	// version
	version();

	if (argc != 3)
	{
		cerr << "Usage: reverse input maxCapacity" << endl;
		exit(1);
	}

	ifstream in(argv[1]);
	float max_capacity = atof(argv[2]);
	string line;
	map<Rule, float> bpMap;
	int fileID = 0;

	while (getline(in, line))
	{
		if (line.find("|||") == string::npos)
		{
			continue;
		}

		// source phrase
		int spp1 = 0,
			spp2 = line.find(" ||| ", spp1);
		string srcPhrase = line.substr(spp1, spp2 - spp1);

		// target phrase
		spp1 = spp2 + 5;
		spp2 = line.find(" ||| ", spp1);
		string trgPhrase = line.substr(spp1, spp2 - spp1);

		// alignment
		spp1 = spp2 + 5;
		spp2 = line.find(" ||| ", spp1);
		map<pair<int, int>, float> alignment;
		istringstream iss(line.substr(spp1, spp2 - spp1).c_str());
		string w;

		while (iss >> w)
		{
			int spp1 = w.find(':'),
			    spp2 = w.find('/'),
				sp = atoi(w.substr(0, spp1).c_str()),
				tp = atoi(w.substr(spp1 + 1, spp2 - spp1 - 1).c_str());
			float prob = atof(w.substr(spp2 + 1, (int)w.size() - spp2 - 1).c_str());

			alignment.insert(map<pair<int, int>, float>::value_type(pair<int, int>(tp, sp), prob));
		}
				
		// frequency
		spp1 = spp2 + 5;
		spp2 = line.size();
		float freq = atof(line.substr(spp1, spp2 - spp1).c_str());

		Rule r;
		r.srcPhrase = trgPhrase;
		r.trgPhrase = srcPhrase;
		r.alignment = alignment;

		// update
		map<Rule, float>::iterator iter = bpMap.find(r);

		if (iter == bpMap.end())
		{
			bpMap.insert(map<Rule, float>::value_type(r, freq));
		}
		else
		{
			iter->second += freq;
		}

		// dump
		if (bpMap.size() > max_capacity)
		{
			fileID++;
			char fileName[100];
			sprintf(fileName, "r_ruleFreq_%d.txt", fileID);
			dump(fileName, bpMap);
			bpMap.clear();
		}
	}

	// dump
	if (!bpMap.empty())
	{
		fileID++;
		char fileName[100];
		sprintf(fileName, "r_ruleFreq_%d.txt", fileID);
		dump(fileName, bpMap);
	}

	return 0;
}
